@@ -34,7 +34,7 @@ static PixmapPtr get_drawable_pixmap(DrawablePtr drawable)
 		return (*drawable->pScreen->GetWindowPixmap)((WindowPtr)drawable);
 }
 
-DRI2BufferPtr
+static DRI2BufferPtr
 nouveau_dri2_create_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, unsigned int attachment,
 			   unsigned int format)
 {
@@ -62,7 +62,10 @@ nouveau_dri2_create_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, unsigned int a
 			ppix->refcnt++;
 	} else {
 		int bpp;
-		unsigned int usage_hint = NOUVEAU_CREATE_PIXMAP_TILED;
+		unsigned int usage_hint = 0;
+
+		if (pNv->Architecture >= NV_ARCH_10)
+			usage_hint |= NOUVEAU_CREATE_PIXMAP_TILED;
 
 		/* 'format' is just depth (or 0, or maybe it depends on the caller) */
 		bpp = round_up_pow2(format ? format : pDraw->depth);
@@ -97,7 +100,7 @@ nouveau_dri2_create_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, unsigned int a
 		nvpix = nouveau_pixmap(ppix);
 		if (!nvpix || !nvpix->bo ||
 		    nouveau_bo_name_get(nvpix->bo, &nvbuf->base.name)) {
-			pScreen->DestroyPixmap(nvbuf->ppix);
+			dixDestroyPixmap(nvbuf->ppix, 0);
 			free(nvbuf);
 			return NULL;
 		}
@@ -105,7 +108,7 @@ nouveau_dri2_create_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, unsigned int a
 	return &nvbuf->base;
 }
 
-DRI2BufferPtr
+static DRI2BufferPtr
 nouveau_dri2_create_buffer(DrawablePtr pDraw, unsigned int attachment,
 			   unsigned int format)
 {
@@ -113,7 +116,7 @@ nouveau_dri2_create_buffer(DrawablePtr pDraw, unsigned int attachment,
 					   attachment, format);
 }
 
-void
+static void
 nouveau_dri2_destroy_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, DRI2BufferPtr buf)
 {
 	struct nouveau_dri2_buffer *nvbuf;
@@ -123,17 +126,17 @@ nouveau_dri2_destroy_buffer2(ScreenPtr pScreen, DrawablePtr pDraw, DRI2BufferPtr
 		return;
 
 	if (nvbuf->ppix)
-	    pScreen->DestroyPixmap(nvbuf->ppix);
+	    dixDestroyPixmap(nvbuf->ppix, 0);
 	free(nvbuf);
 }
 
-void
+static void
 nouveau_dri2_destroy_buffer(DrawablePtr pDraw, DRI2BufferPtr buf)
 {
 	nouveau_dri2_destroy_buffer2(pDraw->pScreen, pDraw, buf);
 }
 
-void
+static void
 nouveau_dri2_copy_region2(ScreenPtr pScreen, DrawablePtr pDraw, RegionPtr pRegion,
 			 DRI2BufferPtr pDstBuffer, DRI2BufferPtr pSrcBuffer)
 {
@@ -211,7 +214,7 @@ nouveau_dri2_copy_region2(ScreenPtr pScreen, DrawablePtr pDraw, RegionPtr pRegio
 	FreeScratchGC(pGC);
 }
 
-void
+static void
 nouveau_dri2_copy_region(DrawablePtr pDraw, RegionPtr pRegion,
 			 DRI2BufferPtr pDstBuffer, DRI2BufferPtr pSrcBuffer)
 {
@@ -245,9 +248,13 @@ static uint64_t dri2_sequence;
 static Bool
 update_front(DrawablePtr draw, DRI2BufferPtr front)
 {
-	int r;
-	PixmapPtr pixmap;
+	ScrnInfoPtr scrn = xf86ScreenToScrn(draw->pScreen);
+	NVPtr pNv = NVPTR(scrn);
 	struct nouveau_dri2_buffer *nvbuf = nouveau_dri2_buffer(front);
+	struct nouveau_bo *pixmap_bo;
+
+	PixmapPtr pixmap;
+	int r;
 
 	if (draw->type == DRAWABLE_PIXMAP)
 		pixmap = (PixmapPtr)draw;
@@ -256,15 +263,23 @@ update_front(DrawablePtr draw, DRI2BufferPtr front)
 
 	pixmap->refcnt++;
 
+	pNv->exa_force_cp = TRUE;
 	exaMoveInPixmap(pixmap);
-	r = nouveau_bo_name_get(nouveau_pixmap_bo(pixmap), &front->name);
+	pNv->exa_force_cp = FALSE;
+	pixmap_bo = nouveau_pixmap_bo(pixmap);
+
+	if (!pixmap_bo)
+		r = -1;
+	else
+		r = nouveau_bo_name_get(pixmap_bo, &front->name);
+
 	if (r) {
-		(*draw->pScreen->DestroyPixmap)(pixmap);
+		dixDestroyPixmap(pixmap, 0);
 		return FALSE;
 	}
 
 	if (nvbuf->ppix)
-		(*draw->pScreen->DestroyPixmap)(nvbuf->ppix);
+		dixDestroyPixmap(nvbuf->ppix, 0);
 
 	front->pitch = pixmap->devKind;
 	front->cpp = pixmap->drawable.bitsPerPixel / 8;
@@ -279,23 +294,27 @@ can_exchange(DrawablePtr draw, PixmapPtr dst_pix, PixmapPtr src_pix)
 	ScrnInfoPtr scrn = xf86ScreenToScrn(draw->pScreen);
 	xf86CrtcConfigPtr xf86_config = XF86_CRTC_CONFIG_PTR(scrn);
 	NVPtr pNv = NVPTR(scrn);
-	int i;
+	int i, active_crtc_count = 0;
 
 	if (!xf86_config->num_crtc)
 		return FALSE;
 
 	for (i = 0; i < xf86_config->num_crtc; i++) {
 		xf86CrtcPtr crtc = xf86_config->crtc[i];
-		if (crtc->enabled && crtc->rotatedData)
-			return FALSE;
+		if (xf86_crtc_on(crtc)) {
+			if (crtc->rotatedData)
+				return FALSE;
 
+			active_crtc_count++;
+		}
 	}
 
 	return ((DRI2CanFlip(draw) && pNv->has_pageflip)) &&
 		dst_pix->drawable.width == src_pix->drawable.width &&
 		dst_pix->drawable.height == src_pix->drawable.height &&
 		dst_pix->drawable.bitsPerPixel == src_pix->drawable.bitsPerPixel &&
-		dst_pix->devKind == src_pix->devKind;
+		dst_pix->devKind == src_pix->devKind &&
+		active_crtc_count;
 }
 
 static Bool
@@ -475,7 +494,7 @@ dri2_page_flip(DrawablePtr draw, PixmapPtr back, void *priv,
 		int head = drmmode_crtc(config->crtc[i]);
 		void *token;
 
-		if (!config->crtc[i]->enabled)
+		if (!xf86_crtc_on(config->crtc[i]))
 			continue;
 
 		flipdata->flip_count++;
@@ -580,8 +599,8 @@ nouveau_wait_vblank(DrawablePtr draw, int type, CARD64 msc,
 	int head;
 
 	/* Select crtc which shows the largest part of the drawable */
-	crtc = nouveau_pick_best_crtc(scrn, FALSE,
-                                  draw->x, draw->y, draw->width, draw->height);
+	crtc = nouveau_pick_best_crtc(scrn,
+				      draw->x, draw->y, draw->width, draw->height);
 
 	if (!crtc) {
 		xf86DrvMsg(scrn->scrnIndex, X_WARNING,
@@ -658,8 +677,8 @@ nouveau_dri2_finish_swap(DrawablePtr draw, unsigned int frame,
 	REGION_TRANSLATE(0, &reg, draw->x, draw->y);
 
 	/* Main crtc for this drawable shall finally deliver pageflip event. */
-	ref_crtc = nouveau_pick_best_crtc(scrn, FALSE, draw->x, draw->y,
-                                      draw->width, draw->height);
+	ref_crtc = nouveau_pick_best_crtc(scrn, draw->x, draw->y,
+					  draw->width, draw->height);
 
 	/* Update frontbuffer pixmap and name: Could have changed due to
 	 * window (un)redirection as part of compositing.
@@ -1020,15 +1039,16 @@ nouveau_dri2_fini(ScreenPtr pScreen)
 }
 
 #ifdef DRI3
-static int is_render_node(int fd, struct stat *st)
+static int is_render_node(int fd)
 {
-	if (fstat(fd, st))
+	struct stat st;
+	if (fstat(fd, &st))
 		return 0;
 
-	if (!S_ISCHR(st->st_mode))
+	if (!S_ISCHR(st.st_mode))
 		return 0;
 
-	return st->st_rdev & 0x80;
+	return st.st_rdev & 0x80;
   }
 
 static int
@@ -1037,7 +1057,6 @@ nouveau_dri3_open(ScreenPtr screen, RRProviderPtr provider, int *out)
 	ScrnInfoPtr pScrn = xf86ScreenToScrn(screen);
 	NVPtr pNv = NVPTR(pScrn);
 	int fd = -1;
-	struct stat buff;
 
 #ifdef O_CLOEXEC
 	fd = open(pNv->render_node, O_RDWR | O_CLOEXEC);
@@ -1047,11 +1066,7 @@ nouveau_dri3_open(ScreenPtr screen, RRProviderPtr provider, int *out)
 	if (fd < 0)
 		return -BadAlloc;
 
-	if (fstat(fd, &buff)) {
-		close(fd);
-		return -BadMatch;
-	}
-	if (!is_render_node(fd, &buff)) {
+	if (!is_render_node(fd)) {
 		drm_magic_t magic;
 
 		if (drmGetMagic(fd, &magic) || drmAuthMagic(pNv->dev->fd, magic)) {
@@ -1072,13 +1087,16 @@ static PixmapPtr nouveau_dri3_pixmap_from_fd(ScreenPtr screen, int fd, CARD16 wi
 	struct nouveau_bo *bo = NULL;
 	struct nouveau_pixmap *nvpix;
 
-	if (depth < 8 || depth > 32 || depth % 8)
+	if (depth < 8 || depth > 32)
 		return NULL;
 
 	pixmap = screen->CreatePixmap(screen, 0, 0, depth, 0);
 	if (!pixmap)
 		return NULL;
 
+	if (pixmap->drawable.bitsPerPixel % 8)
+		goto free_pixmap;
+
 	if (!screen->ModifyPixmapHeader(pixmap, width, height, 0, 0, stride, NULL))
 		goto free_pixmap;
 
@@ -1092,7 +1110,7 @@ static PixmapPtr nouveau_dri3_pixmap_from_fd(ScreenPtr screen, int fd, CARD16 wi
 	return pixmap;
 
 free_pixmap:
-	screen->DestroyPixmap(pixmap);
+	dixDestroyPixmap(pixmap, 0);
 	return NULL;
 }
 
@@ -1124,15 +1142,13 @@ nouveau_dri3_screen_init(ScreenPtr screen)
 #ifdef DRI3
 	ScrnInfoPtr pScrn = xf86ScreenToScrn(screen);
 	NVPtr pNv = NVPTR(pScrn);
-	struct stat master, render;
 	char *buf;
 
-	if (is_render_node(pNv->dev->fd, &master))
+	if (is_render_node(pNv->dev->fd))
 		return TRUE;
 
 	buf = drmGetRenderDeviceNameFromFd(pNv->dev->fd);
-	if (buf && stat(buf, &render) == 0 &&
-	    master.st_mode == render.st_mode) {
+	if (buf) {
 		pNv->render_node = buf;
 		if (dri3_screen_init(screen, &nouveau_dri3_screen_info)) {
 			xf86DrvMsg(pScrn->scrnIndex, X_INFO,
